=encoding utf8

=cut

package Text::Compose;

use warnings;
use strict;
use utf8;

=head1 NAME

Text::Compose - Compose decomposable script files

=head1 VERSION

Version 0.15

=cut

our $VERSION = '0.15';

=head1 SYNOPSIS

	use Text::Compose;
	use Data::Dumper;

	my $block;
	my $buf = "";

	$block = Text::Compose::block("Frittenbude");
	$block->text("Meine\nFrittenbude\n");
	$block->value("meine", 1);
	$block->value("wirklich", "Ja doch");
	$block->text("Bestelle ");
	$block->param("basis", "pommes");
	$block->text(" mit ");
	$block->param("sosse", "mayo");
	$block->text(".\n");
	$buf .= $block->result();
	print $buf;

	my %result;

	%result = Text::Compose::decompose_lines($buf);
	print Dumper(\%result);

=head1 DESCRIPTION

Text::Compose allows the composition of decomposable script files.
That is, an automatically generated script (or other text) file including some variable components is interspersed with pseudo-comments that allow those variable components to be re-extracted als a Perl hash.
For example, composing a portion of a Sieve script with

	use Text::Compose;

	[...]

	my $script = ...;
	my $address = ...;
	my $keep = ...;
	my $ref = ref $address ? $address : [ $address ];
	my $block = Text::Compose::block("redirect");
	foreach my $addr (@{$ref}) {
		$block->text("redirect \"");
		$block->param("address", $addr);
		$block->text("\";\n");
	}
	if ($keep) {
		$block->value("keep", 1);
		$block->text("keep;\n");
	}
	$script .= $block->result();

gives you a working Sieve script that later can be decomposed: Calling

	my $input = ...;
	Text::Compose::decompose_file($input);

will return a Perl hash containing an element "redirect" whose "address" and "keep" fields reflect the values used in composing the script.

=head1 EXPORT

block -- start composing a new block
result -- return result of block composition
text -- paste text into block
value -- record a parameter value not present in the block
param -- record a parameter value present in the block text

decompose_array -- decompose from an array of lines
decompose_lines -- decompose from a string of lines
decompose_file -- decompose from a file
decompose_error -- error of last unsuccessful decompose

=head1 DEPENDENCIES

Text::Compose needs the following modules:

=over

=item * Encode

=item * Digest::MD5

=back

=cut

use Encode;
use Digest::MD5 qw/md5_base64/;

#use Data::Dumper;

=head1 FUNCTIONS

=cut

=for comment
Private functions

=cut

sub _addvalue {
	use strict 'refs';
	my ($hash_ref, $key, $value) = @_;
	if (!exists $hash_ref->{$key}) {
		$hash_ref->{$key} = $value;
	} elsif (ref($hash_ref->{$key}) eq "") {
		$hash_ref->{$key} = [$hash_ref->{$key}, $value];
	} elsif (ref($hash_ref->{$key}) eq "ARRAY") {
		push @{$hash_ref->{$key}}, $value;
	} else {
	}
}

sub _encode {
	my ($self, $text, @encodings) = @_;
#print "ENCODE: $text";
	foreach my $encoding (@encodings) {
		if ($encoding eq "utf-8") {
			$text = Encode::encode("utf-8", $text);
		} elsif ($encoding eq "encoded-character") {
			$text =~ s/([\$\P{IsASCII}])/sprintf '${Unicode:%X}', ord($1)/ge;
		} elsif ($encoding eq "dot-stuff") {
			$text =~ s/^\./../mg;
		} elsif ($encoding eq "quote-esc") {
			$text =~ s/(["\\])/\\$1/g;
		} else {
		}
	}
#print "-> $text\n";
	return $text;
}

sub _decode {
	my ($text, @encodings) = @_;
#print "DECODE(", join(",", @encodings), "): $text";
	foreach my $encoding (reverse @encodings) {
		if ($encoding eq "utf-8") {
			$text = decode("utf-8", $text);
		} elsif ($encoding eq "encoded-character") {
			$text =~ s/\$\{Unicode: *([\p{IsXDigit}]+)\}/chr(hex($1))/ge;
		} elsif ($encoding eq "dot-stuff") {
			$text =~ s/^\.\././mg;
		} elsif ($encoding eq "quote-esc") {
			$text =~ s/\\(["\\])/$1/g;
		} else {
		}
	}
#print "-> $text\n";
	return $text;
}

=head2 block(blockname, comment_begin, comment_end, encoding);

Begins a new block.

=head3 blockname

Name of the block. This will be the top-level hash key upon decomposition.

=head3 comment_begin, comment_end

Strings that denote the beginning and end of a comment in the language used.
Defaults are "# " and the empty string, respectively.

=head3 encoding

Optional default text encoding(s) to use for parameter values.

=cut

sub block {
	my ($blockname, $comment_begin, $comment_end, $encoding) = @_;
	my %self;
	$self{blockname} = $blockname;
	$self{comment_begin} = $comment_begin || "# ";
	$self{comment_end} = $comment_end || "";
	$self{buf} = ""; $self{row} = 0; $self{col} = 0;
	$self{params} = "";
	if ($encoding) {
		if (ref $encoding eq "ARRAY") {
			$self{encodings} = $encoding;
		} elsif (ref $encoding) {
		} else {
			$self{encodings} = [$encoding];
		}
	}
	return bless \%self;
}

=head2 result(block)

Ends a block and returns the result.

=head3 block

Block reference as returned by C<block>.

=cut

sub result {
	my ($self) = @_;
	if ($self->{col} > 0) {
		$self->{buf} .= "\n";
		$self->{row}++; $self->{col} = 0;
	}
#print Dumper($self->{params}, $self->{buf});
	utf8::upgrade($self->{params}); utf8::upgrade($self->{buf});
#print Dumper($self->{params}, $self->{buf});
	my $md5 = md5_base64($self->{params}, $self->{buf});
#print "MD5: $md5\n";
	my $header = $self->{comment_begin} . $self->{blockname} . ":" . $self->{row} . "*" . $md5 . " " . $self->{params} . $self->{comment_end} . "\n";
	return $header . $self->{buf};
}

=head2 text(block, text)

Write verbatim text into the block.

=head3 block

Block reference as returned by C<block>.

=head3 text

Text to place into the block.

=cut

sub text {
	my ($self, $text) = @_;
	my @matches;
	@matches = $text =~ m/\n/go;
	if (@matches) {
		$self->{row} += @matches;
		$self->{col} = defined($') ? length($') : 0;
	} else {
		$self->{col} += length($text);
	}	
	$self->{buf} .= $text;
}

=head2 value(block, name, value)

Note the value of a parameter without actually writing into the non-comment part of a bloc.

=head3 block

Block reference as returned by C<block>.

=head3 name, value

Name and value of the parameter. C<name> will be the second-level hash key upon decomposition.

=cut

sub value {
	my ($self, $name, $value) = @_;
	$self->{params} .= " " if $self->{params};
	$self->{params} .= $name;
	if (index($value, " ") >= 0) {
		$self->{params} .= "#" . length($value) . ":" . $value;
	} else {
		$self->{params} .= "=" . $value;
	}
}

=head2 param(block, name, value, encoding)

Write a named parameter value into the block and note its position for later retrieval (decomposition).

=head3 block

Block reference as returned by C<block>.

=head3 name, value

Name and value of the parameter. C<name> will be the second-level hash key upon decomposition.

=head3 encoding

Optional text encoding to use. Defaults to the C<encoding> parameter of C<block>.

=cut

sub param {
	my ($self, $name, $value, $encoding) = @_;
	$self->{params} .= " " if $self->{params};
	$self->{params} .= $name . "@" . ($self->{row} + 1) . "," . ($self->{col} + 1);
	my @encodings;
	@encodings = $self->{encodings} if defined $self->{encodings};
	if ($encoding) {
		if (ref $encoding eq "ARRAY") {
			unshift @encodings, @{$encoding};
		} elsif (ref $encoding) {
		} else {
			unshift @encodings, $encoding;
		}
	}
#print Dumper(\@encodings);
	$value = $self->_encode($value, @encodings) if @encodings;
	$self->text($value);
	$self->{params} .= "-" . ($self->{row} + 1) . "," . ($self->{col} + 1);
	$self->{params} .= "/" . join ",", @encodings if @encodings;
}

sub _decompose_header {
	my ($header, $comment_begin, $comment_end) = @_;
	$comment_begin ||= "# "; $comment_end ||= "";
#print "HEADER: $header\n";
	my ($name, $lines, $md5, $params);
	$header =~ m/$comment_begin(\w+):(\d+)(\*([A-Za-z0-9+\/]*))? (.*)$comment_end/;
	$name = $1; $lines = $2; $md5 = $4; $params = $5;
#print "NAME: $name, LINES: $lines, MD5: $md5, PARAMS: $params\n";
	return ($name, $lines, $md5, $params);
}

=for comment
Private functions

=cut

sub _decompose_body {
	my ($params, $body_ref) = @_;
#print "BODY: " . Dumper($body_ref);
	my %param;
	my ($name, $value);
#print "PARAMS: $params\n";
	while ($params =~ m/\G ?(\w+)([=#@])/g) {
		$name = $1;
		my $method = $2;
#print "NAME: $name, METHOD: $method\n";
		if ($method eq "=") {
			$params =~ m/\G([^ ])/g;
			$value = $1;
#print "IMMED: $value\n";
		} elsif ($method eq "#") {
			$params =~ m/\G(\d+):/g;
			my $length = $1;
			$params =~ m/\G(.{$length})/g;
			$value = $1;
#print "LEN $length: $value\n";
		} elsif ($method eq "@") {
			my ($begin_row, $begin_col, $end_row, $end_col);
			$params =~ m/\G(\d+),(\d+)-(\d+),(\d+)/g;
			$begin_row = $1 - 1; $begin_col = $2 - 1; $end_row = $3 - 1; $end_col = $4 - 1;
			if ($begin_row == $end_row) {
#print "SAME: $body_ref->[$begin_row] $begin_col-$end_col\n";
				$value = substr($body_ref->[$begin_row], $begin_col, $end_col - $begin_col);
			} else {
				$value = substr($body_ref->[$begin_row], $begin_col) . "\n";
#print "FIRST: $value";
				$value .= $body_ref->[$_] . "\n" foreach ($begin_row + 1 .. $end_row - 1);
#print "LOOP: $value";
				$value .= substr($body_ref->[$end_row], 0, $end_col);
#print "LAST: $value";
			}
#print "POS $begin_row,$begin_col-$end_row,$end_col: $value\n";
		}
		if ($params =~ m/\G\/([-,\w]+)/gc) {
#print "ENCODING: $1\n";
			$value = _decode($value, split /,/, $1);
#print "DECODED: $value\n";
		}
		_addvalue(\%param, $name, $value);
	}
#print Dumper(\%param);
	return \%param;
}

=for comment
Private variable

=cut

my $_decompose_error;

=head2 decompose_error()

Returns an error message describing why the last decompose_xxx failed (if it failed, that is).

=cut

sub decompose_error {
	return $_decompose_error;
}

=head2 decompose_array(arr_ref)

Decompose a sequence of blocks given as an array of lines. Returns a hash of hashes, first level keys being the names of the individual blocks and second level keys being parameter names.

=head3 arr_ref

Reference to an array of lines to decompose.

=cut

sub decompose_array {
	my ($arr_ref) = @_;
#print "ARRAY: " . Dumper($arr_ref);
	my $offset = 0;
	my ($name, $lines, $md5, $params);
	my %result;
#print "LEN: " . (0 + @{$arr_ref}) . "\n";
	while ($offset < @{$arr_ref}) {
#print "OFFSET: $offset\n";
#print "HEADER: $arr_ref->[$offset]\n";
		($name, $lines, $md5, $params) = _decompose_header(@{$arr_ref}[$offset]);
#print "NAME: $name, LINES: $lines, PARAMS: $params\n";
#print "SLICE: $offset + 1 .. $offset + $lines\n";
		my @body = @{$arr_ref}[$offset + 1 .. $offset + $lines];
#print "BODY: ", Dumper(\@body);
		if ($md5) {
			utf8::upgrade($params);
			utf8::upgrade($_) foreach @body;
			my $md5_act = md5_base64($params, join("\n", @body), "\n");
#print "MD5 <- ", Dumper($params, join("\n", @body), "\n");
#print "MD5_ACT: $md5_act\n";
			if ($md5_act ne $md5) {
				$_decompose_error = "checksum mismatch";
				return ();
			}
		}
#print "BODY: " . Dumper(\@body);
		$result{$name} = _decompose_body($params, \@body);
#		$result{$name} = _decompose_body($params, \@{$arr_ref}[$offset + 1 .. $offset + $lines]);
		$offset += 1 + $lines;
	}
	undef $_decompose_error;
	return %result;
}

=head2 decompose_lines(lines)

Decompose a sequence of blocks in a string of lines. Returns a hash of hashes, first level keys being the names of the individual blocks and second level keys being parameter names.

=head3 lines

String of lines to decompose.

=cut

sub decompose_lines {
	my ($lines) = @_;
	$lines .= "."; # Leerzeilen am Ende gehen sonst verloren
	my @arr = split /\n/,$lines;
	delete $arr[$#arr]; # Anhängsel wieder abschneiden
	return decompose_array(\@arr);
}

=head2 decompose_files(handle)

Decompose a sequence of blocks in a file. Returns a hash of hashes, first level keys being the names of the individual blocks and second level keys being parameter names.

=head3 handle

File handle of an open file to read.

=cut

sub decompose_file {
	my ($handle) = @_;
	my $line;
	my ($name, $lines, $md5, $params);
	my %result;
	for (;;) {
		undef $!;
		if (!defined($line = readline($handle))) {
			if ($!) {
				$_decompose_error = "readline: $!";
				return ();
			}
			last;
		}
		chomp $line;
#print "HEADER: $line\n";
		($name, $lines, $md5, $params) = _decompose_header($line);
		my @body;
		foreach my $i (1 .. $lines) {
			undef $!;
			if (!defined($line = readline($handle))) {
				if ($!) {
					$_decompose_error = "readline: $!";
					return ();
				}
				$_decompose_error = "unexpected EoF";
				return ();
			}
			push @body, $line
		}
		if ($md5) {
			utf8::upgrade($params);
			utf8::upgrade($_) foreach @body;
			my $md5_act = md5_base64($params, @body);
#print "MD5_ACT: $md5_act\n";
			if ($md5_act ne $md5) {
				$_decompose_error = "checksum mismatch";
				return ();
			}
		}
		chomp @body;
#print Dumper(\@body);
		$result{$name} = _decompose_body($params, \@body);
	}
	undef $_decompose_error;
	return %result;
}

=head1 AUTHOR

Edgar Fuß, Mathematisches Institut der Universität Bonn, C<< <ef at math.uni-bonn.de> >>

=head1 BUGS

Please report any bugs or feature requests to C<bug-text-compose at rt.cpan.org>, or through
the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=Text-Compose>.  I will be notified, and then you'll
automatically be notified of progress on your bug as I make changes.


=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Text::Compose


You can also look for information at:

=over

=item * RT: CPAN's request tracker

L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=Text-Compose>

=item * AnnoCPAN: Annotated CPAN documentation

L<http://annocpan.org/dist/Text-Compose>

=item * CPAN Ratings

L<http://cpanratings.perl.org/d/Text-Compose>

=item * Search CPAN

L<http://search.cpan.org/dist/Text-Compose/>

=back

=head1 COPYRIGHT & LICENSE

Copyright 2010 Edgar Fuß, Mathematisches Institut der Universität Bonn.

This program is free software; you can redistribute it and/or modify it
under the terms of the Artistic License.

See http://dev.perl.org/licenses/ for more information.

=cut

1; # End of Text::Compose
